"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Sql = require("sequelize");
const _ = require("lodash");
const transaction_decorator_1 = require("../../decorators/orm/transaction.decorator");
const prop_1 = require("../../metadata/orm/prop");
const model_1 = require("../../metadata/orm/model");
const belongs_to_1 = require("../../metadata/orm/associations/belongs-to");
const has_one_1 = require("../../metadata/orm/associations/has-one");
const orm_1 = require("../../decorators/orm");
const transaction_impl_1 = require("./transaction-impl");
class DbImpl {
    constructor(modelFn, model, sequelize, logger, transactionService) {
        this.modelFn = modelFn;
        this.model = model;
        this.sequelize = sequelize;
        this.logger = logger;
        this.transactionService = transactionService;
        this.createCopyValsFn();
        this.createTransformQuery();
    }
    getSqlTransact(transaction) {
        transaction = transaction || this.transactionService.current;
        if (transaction)
            this.logger.verbose('dbimpl', `Using transaction: ${transaction.fullName}`);
        return transaction && transaction.sync();
    }
    transaction(name, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.sequelize.transaction(name, transaction);
        });
    }
    create(t, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            if (t instanceof Array) {
                t = _.cloneDeep(t);
                t.forEach((t, idx, arr) => { arr[idx] = this.transformQueryWhere(t); });
                let results = yield this.model.bulkCreate(t, _.merge({}, { transaction: sqlTransact }));
                return true;
            }
            else {
                t = this.transformQueryWhere(t);
                let result = yield this.model.create(t, _.merge({}, { transaction: sqlTransact }));
                return this.wrapResult(result);
            }
        });
    }
    findById(id, options, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            if (options)
                options = this.transformQuery(options);
            let result = yield this.model.findById(id, _.merge({}, { transaction: sqlTransact }, options));
            return result && this.wrapResult(result);
        });
    }
    findOne(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            query = this.transformQuery(query);
            let result = yield this.model.findOne(_.merge({}, { transaction: sqlTransact }, query));
            return result && this.wrapResult(result);
        });
    }
    findOrCreate(query, defaults, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            query = this.transformQueryWhere(query);
            defaults = this.transformQueryWhere(defaults);
            let [result, created] = yield this.model.findOrCreate(_.merge({}, { where: query, defaults: defaults || {}, transaction: sqlTransact }));
            return [result && this.wrapResult(result), created];
        });
    }
    findAndCountAll(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            if (query)
                query = this.transformQuery(query);
            let results = yield this.model.findAndCountAll(_.merge({}, { transaction: sqlTransact }, query));
            return { count: results.count, results: this.wrapResults(results.rows) };
        });
    }
    findAll(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            if (query)
                query = this.transformQuery(query);
            let results = yield this.model.findAll(_.merge({}, { transaction: sqlTransact }, query));
            return this.wrapResults(results);
        });
    }
    all(query, transaction) {
        return this.findAll(query, transaction);
    }
    count(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            if (query)
                query = this.transformQuery(query);
            return yield this.model.count(_.merge({}, { transaction: sqlTransact }, query));
        });
    }
    max(field, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            return yield this.model.max(field, _.merge({}, { transaction: sqlTransact }));
        });
    }
    min(field, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            return yield this.model.min(field, _.merge({}, { transaction: sqlTransact }));
        });
    }
    sum(field, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            return yield this.model.sum(field, _.merge({}, { transaction: sqlTransact }));
        });
    }
    save(t, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate({ id: t.id }, t, transaction);
            return result;
        });
    }
    update(query, replace, returning = false, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!query) {
                throw new Error(`Db.update query parameter was falsey: ${query}`);
            }
            let sqlTransact = this.getSqlTransact(transaction);
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else {
                query = this.transformQuery(query);
                if (returning) {
                    let results = yield this.model.findAll(query);
                    let ids = results.map((result) => result.id);
                    query = { where: { id: { $in: ids } } };
                }
            }
            replace = this.transformQueryWhere(replace);
            let [affected, results] = yield this.model.update(replace, _.merge({}, { transaction: sqlTransact }, query));
            if (returning) {
                results = this.wrapResults(yield this.model.findAll(_.merge({}, { transaction: sqlTransact }, query)));
            }
            return [affected, results];
        });
    }
    updateOrCreate(query, defaults, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate(query, defaults);
            if (!created) {
                let worked = yield this.update({ where: query }, defaults, false);
                if (!worked)
                    throw new Error("Failed to update or create a model.");
                let resultOrNull = yield this.findOne(_.merge({}, { where: query }, defaults));
                if (!resultOrNull)
                    throw new Error("Updated row, but could not find it afterwards.");
                result = resultOrNull;
            }
            return [result, created];
        });
    }
    destroy(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = this.getSqlTransact(transaction);
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else
                query = this.transformQuery(query);
            return yield this.model.destroy(_.merge({}, { transaction: sqlTransact }, query));
        });
    }
    isId(query) {
        return typeof query === 'number' || typeof query == 'string';
    }
    isT(query) {
        return !!query.id;
    }
    createCopyValsFn() {
        let allProps = [];
        let directTransformFn = (val) => val;
        let props = Reflect.getOwnMetadata(model_1.ModelPropertiesSym, this.modelFn.prototype) || [];
        for (let q = 0; q < props.length; q++) {
            let propName = props[q];
            let propMeta = Reflect.getOwnMetadata(prop_1.PropMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model property metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let transformFn = directTransformFn;
            if (propMeta.type == orm_1.Types.DATE)
                transformFn = (dateStr) => new Date(dateStr);
            allProps.push({ columnName: propMeta.columnName || propName, propertyName: propName, transformFn: transformFn });
        }
        this.copyVals = function (sql, t) {
            for (let q = 0; q < allProps.length; q++) {
                let propName = allProps[q].propertyName;
                t[propName] = allProps[q].transformFn(sql[propName]);
            }
        };
    }
    static getForeignModelDbImpl(foreignModel) {
        if (!foreignModel || !("db" in foreignModel))
            throw new Error("Cannot get DbImpl from ForeignModel that has not been resolved");
        let staticModel = foreignModel;
        return staticModel.db;
    }
    createTransformQuery() {
        let transforms = this.getTransforms();
        this.createTransformQueryWhere(transforms);
        this.createTransformQueryInclude(transforms);
        this.createTransformResult(transforms);
    }
    transformQuery(query) {
        let result = _.clone(query);
        if (query.where)
            result.where = this.transformQueryWhere(query.where);
        if (query.include)
            result.include = this.transformQueryInclude(query.include);
        return result;
    }
    getTransforms() {
        let transforms = [];
        let belongsTo = Reflect.getOwnMetadata(belongs_to_1.ModelBelongsToAssociationsSym, this.modelFn.prototype) || [];
        for (let q = 0; q < belongsTo.length; q++) {
            let propName = belongsTo[q];
            let propMeta = Reflect.getOwnMetadata(belongs_to_1.BelongsToMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model belongs-to metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for belongs-to property ${this.modelFn.name || this.modelFn}.${propName}`);
            transforms.push({
                type: 'belongs-to',
                columnName: fkey,
                fieldName: propName,
                foreignPkName: 'id',
                foreignDb: () => DbImpl.getForeignModelDbImpl(propMeta.foreignModel)
            });
        }
        let hasOne = Reflect.getOwnMetadata(has_one_1.ModelHasOneAssociationsSym, this.modelFn.prototype) || [];
        for (let q = 0; q < hasOne.length; q++) {
            let propName = hasOne[q];
            let propMeta = Reflect.getOwnMetadata(has_one_1.HasOneMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model has-one metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for has-one property ${this.modelFn.name || this.modelFn}.${propName}`);
            transforms.push({
                type: 'has-one',
                foreignColumnName: fkey,
                fieldName: propName,
                pkName: 'id',
                foreignDb: () => DbImpl.getForeignModelDbImpl(propMeta.foreignModel)
            });
        }
        return transforms;
    }
    createTransformQueryWhere(transforms) {
        this.transformQueryWhere = function (query) {
            if (!query)
                return query;
            query = _.clone(query);
            if (query['$and']) {
                let andVal = query['$and'];
                if (Array.isArray(andVal)) {
                    for (let q = 0; q < andVal.length; q++) {
                        if (typeof andVal[q] !== 'object')
                            throw new Error(`Invalid $and query: [..., ${andVal[q]}, ...]`);
                        andVal[q] = this.transformQueryWhere(andVal[q]);
                    }
                }
                else if (typeof andVal === 'object')
                    andVal = this.transformQueryWhere(andVal);
                else
                    throw new Error(`Invalid $and query: ${andVal}`);
                query['$and'] = andVal;
            }
            if (query['$or']) {
                let orVal = query['$or'];
                if (Array.isArray(orVal)) {
                    for (let q = 0; q < orVal.length; q++) {
                        if (typeof orVal[q] !== 'object')
                            throw new Error(`Invalid $or query: [..., ${orVal[q]}, ...]`);
                        orVal[q] = this.transformQueryWhere(orVal[q]);
                    }
                }
                else if (typeof orVal === 'object')
                    orVal = this.transformQueryWhere(orVal);
                else
                    throw new Error(`Invalid $or query: ${orVal}`);
                query['$or'] = orVal;
            }
            for (let q = 0; q < transforms.length; q++) {
                let transform = transforms[q];
                let fieldVal = query[transform.fieldName];
                switch (transform.type) {
                    case 'belongs-to':
                        if (typeof fieldVal !== 'undefined') {
                            if (fieldVal && fieldVal[transform.foreignPkName])
                                fieldVal = fieldVal[transform.foreignPkName];
                            query[transform.columnName] = fieldVal;
                            delete query[transform.fieldName];
                        }
                        break;
                    case 'has-one':
                        if (typeof fieldVal !== 'undefined') {
                            throw new Error(`Not implemented! Cannot include has-one value in where query`);
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return query;
        };
    }
    createTransformQueryInclude(transforms) {
        function getFieldModel(field) {
            for (let i = 0; i < transforms.length; i++) {
                if (field == transforms[i].fieldName)
                    return transforms[i].foreignDb().model;
            }
            return null;
        }
        this.transformQueryInclude = function (fields) {
            if (!fields)
                return fields;
            return fields.map(field => {
                let model = getFieldModel(field);
                if (!model)
                    throw new Error(`Cannot find field ${field} from include query`);
                return { model: model, as: field };
            });
        };
    }
    createTransformResult(transforms) {
        this.transformResult = function (sql, result) {
            if (result === null)
                return result;
            else if (typeof result === 'undefined')
                throw new Error(`Result was undefined in DbImpl#transformResult!`);
            result = _.clone(result);
            for (let q = 0; q < transforms.length; q++) {
                let transform = transforms[q];
                let foreignDb = transform.foreignDb();
                let fieldVal = sql[transform.fieldName];
                switch (transform.type) {
                    case 'belongs-to':
                        if (typeof fieldVal !== 'undefined') {
                            let t = null;
                            if (fieldVal) {
                                t = new foreignDb.modelFn();
                                foreignDb.copyVals(fieldVal, t);
                            }
                            result[transform.fieldName] = transform.foreignDb().transformResult(fieldVal, t);
                        }
                        else if (typeof sql[transform.columnName] !== 'undefined') {
                            result[transform.fieldName] = sql[transform.columnName];
                            delete result[transform.columnName];
                        }
                        break;
                    case 'has-one':
                        if (typeof fieldVal !== 'undefined') {
                            let t = null;
                            if (fieldVal) {
                                t = new foreignDb.modelFn();
                                foreignDb.copyVals(fieldVal, t);
                            }
                            result[transform.fieldName] = foreignDb.transformResult(fieldVal, t);
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return result;
        };
    }
    fromJson(json) {
        return this.wrapResult(json);
    }
    wrapResult(result) {
        let t = new this.modelFn();
        this.copyVals(result, t);
        return this.transformResult(result, t);
    }
    wrapResults(results) {
        return results.map(result => this.wrapResult(result));
    }
}
__decorate([
    transaction_decorator_1.Transaction(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Object, transaction_impl_1.TransactionImpl]),
    __metadata("design:returntype", Promise)
], DbImpl.prototype, "updateOrCreate", null);
exports.DbImpl = DbImpl;
//# sourceMappingURL=db-impl.js.map